#include "stdafx.h"
#include <gl.h>
#include <glu.h>
#include "glmodel.h"
#include "swf.h"

//-----------------------------------------------------------------------------
//-- Interface to .qmf file 
const COLORREF g_aStateColors[10] = { 
	RGB( 0, 0, 0), RGB( 255, 0, 0), 
	RGB( 0, 0, 128), RGB( 0, 128, 0), 
	RGB( 255, 255, 0), RGB( 255, 0, 255), 
	RGB( 0, 255, 0), RGB( 0, 0, 255), 
	RGB( 0, 255, 255), RGB( 0, 255, 255) 
	};

/*

Note : The following functions retrieve information related to glViewport:
	glGet with argument GL_VIEWPORT
	glGet with argument GL_MAX_VIEWPORT_DIMS

ToDo : The Setlevels function calculates the low and high Y values which are 
linearly interpolated between, but this is not a perfect abstraction - recalc at 
all intermediate levels would be preferable.

*/

//----- ID's for manually created controls
const int ID_EDTSTATE0	= 5000;	
const int ID_EDTSCALING	= 5001;
const int ID_EDTMAXSTIM = 5002;
const int ID_CHKGRID	= 5003;
const int ID_CHKSTATES	= 5004;
const int ID_CHKLEGEND	= 5005;
const int ID_CHKCOLOR	= 5006;
const int ID_CHKSTIMULUS= 5007;
const int ID_SCRSTIMULUS= 5008;
const int ID_BTNRESET	= 5009;
const int ID_BTNCOPYIMG	= 5010;
const int ID_CHKROTATEX	= 5011;
const int ID_CHKROTATEY	= 5012;
const int ID_CHKROTATEZ	= 5013;
const int ID_BTNEXPORT	= 5014;
const int ID_CHKFINE	= 5015;
const int ID_EDTYMAX	= 5016;	
const int ID_CHKYMAX	= 5017;	

//----- Location of GL window : MRG from left, MRG from bottom, MRG from right, TOP from top.
const int TOPOFGLWINDOW = 110;		
const int GLWINDOWMARGIN = 3;		

//----- TIMER ID for movies. (whatever value you like )
const int TIMER_ROTATE = 128;
const int TIMER_STIMULUS = 99;
const double ROTATEANGLE = 4.5;

const int STIMULUS_INCREMENT = 5;

const double MINSTATE = -100.0;
const double MAXRATE = 100.0;
const double STATEDESCENT = 10.0;   // Amount below diagram to place states

const GLfloat BLOCKSIZE=6.0f;
const GLfloat BLOCKINDENT=0.1f;
const GLfloat BLOCKTEXTSEP=0.1f;

extern int g_iXDrag;
extern int g_iYDrag;
extern bool g_bDrag;

//-----------------------------------------------------------------------------
class CEnergyModel : public CGLModel { 
	protected : 
		int m_iStimulus;			// ranges from 0..100
		int m_iStimulusDir;			// 1=rising -1=falling
		
	public : 
		CModel m_oModel;

		int m_iStates;
		int m_iRates;
		double m_dEnergyMin, m_dEnergyMax;

		CEnergyModel() { 
			m_iStimulus=0;
			m_iStimulusDir=1;
			}
			
		//----- Set the height of the energy barriers 
		void SetLevels( int iZeroState, double dMaxStim, double dRefRate, double dYMax ) {
			int iRate, iState;
			int iStates = m_oModel.m_aStates.Count();
			int iRates = m_oModel.m_aRates.Count();
			double k0_f, k0_b, k1_f, k1_b;
			double dLnK0 = log( max( 1.0, dRefRate ) );

			//----- Check if model has any P or Q dependance
			bool bPDependent = m_oModel.HasPreExpDepend();
			bool bQDependent = m_oModel.HasExpDepend();

			//----- Initialize array of state levels
			for( iState=0; iState<m_oModel.m_aStates.Count(); ++iState )
				m_oModel.State(iState)->m_dEnergyMin = m_oModel.State(iState)->m_dEnergyMax = MINSTATE-1.0f;
				
			//----- Initialize array of rate (barrier) levels
			for( iRate=0; iRate<m_oModel.m_aRates.Count(); ++iRate ) 
				m_oModel.Rate(iRate)->m_dEnergyMin = m_oModel.Rate(iRate)->m_dEnergyMax = MAXRATE+1.0f;
			
			//----- Set zero state levels
			if( iZeroState<0 || iZeroState>=m_oModel.m_aStates.Count() ) 
				iZeroState = 0;
			m_oModel.State( iZeroState )->m_dEnergyMax = m_oModel.State( iZeroState )->m_dEnergyMin = 0.0f;

			//----- Set Rate levels until all have been set or none are found in current iteration 
			CModelRate * pRate;
			CModelState * pState0, * pState1;
			int iPrev=-1, iSet=0;
			while( iSet<m_oModel.m_aRates.Count() && iPrev<iSet ) {
				iPrev=iSet;
				for( int iRate=0; iRate<m_oModel.m_aRates.Count(); ++iRate ) {
					pRate = m_oModel.Rate(iRate);
					
					if( pRate->m_dEnergyMin > MAXRATE ) { 
						//-- calc k's for min and max stimulus levels
						k0_f = pRate->m_adK0[0];
						k0_b = pRate->m_adK0[1];
						
						if( bQDependent ) { 
double dTemp = 	exp( pRate->m_adK1[0] ) ; 
							k1_f = k0_f * ( pRate->m_abQDep[0] ?  dTemp * dMaxStim : 1.0 );
							k1_b = k0_b * ( pRate->m_abQDep[1] ?  exp( pRate->m_adK1[1] ) * dMaxStim : 1.0 );
							}
						else if( bPDependent ) { 
							k1_f = k0_f * ( pRate->m_abPDep[0] ? dMaxStim : 1.0 );
							k1_b = k0_b * ( pRate->m_abPDep[1] ? dMaxStim : 1.0 );
							}
						else { 
							k1_f = k0_f ;
							k1_b = k0_b ;
							}
							
						//double dLnK0 = log(max(1.0f,dRefRate));

						//-- check if states or rate need to be set 
						pState0 = m_oModel.State( pRate->m_aiStates[0] );
						pState1 = m_oModel.State( pRate->m_aiStates[1] );
						if( pState0->m_dEnergyMin>=MINSTATE && pState1->m_dEnergyMin<MINSTATE ){
							//-- 1 set, 2 not - set rate and 2
							pState1->m_dEnergyMin = pState0->m_dEnergyMin + log( k0_b / k0_f );
							pRate->m_dEnergyMin = pState0->m_dEnergyMin + dLnK0 - log( k0_f );
							pState1->m_dEnergyMax = pState0->m_dEnergyMax + log( k1_b / k1_f );
							pRate->m_dEnergyMax = pState0->m_dEnergyMax + dLnK0 - log( k1_f );
							++iSet;
							}
						else if( pState0->m_dEnergyMin<MINSTATE && pState1->m_dEnergyMin>=MINSTATE ) {
							//-- 2 set, 1 not - set rate and 1
							pState0->m_dEnergyMin = pState1->m_dEnergyMin + log( k0_f / k0_b );
							pRate->m_dEnergyMin = pState1->m_dEnergyMin + dLnK0 - log( k0_b );
							pState0->m_dEnergyMax = pState1->m_dEnergyMax + log( k1_f / k1_b );
							pRate->m_dEnergyMax = pState1->m_dEnergyMax + dLnK0 - log( k1_b );
							++iSet;
							}
						else if( pState0->m_dEnergyMin>=MINSTATE && pState1->m_dEnergyMin>=MINSTATE ) {
							//-- 1 and 2 set, set rate
							pRate->m_dEnergyMin = pState0->m_dEnergyMin + dLnK0 - log( k0_f );
							pRate->m_dEnergyMax = pState0->m_dEnergyMax + dLnK0 - log( k1_f );
							++iSet;
							}
						}       
					}
				}

			//----- Find Y axis limits
			if( dYMax<0 ) { 
				m_dEnergyMin = -1.0;
				for( iState=0; iState<m_oModel.m_aStates.Count(); ++iState )
					m_dEnergyMin = min( min( m_dEnergyMin, m_oModel.State(iState)->m_dEnergyMin ), m_oModel.State(iState)->m_dEnergyMax );
				
				m_dEnergyMax = 1.0;
				for( iRate=0; iRate<m_oModel.m_aRates.Count(); ++iRate ) 
					m_dEnergyMax = max( max( m_dEnergyMax, m_oModel.Rate(iRate)->m_dEnergyMin), m_oModel.Rate(iRate)->m_dEnergyMax );
				}
			else { 
				m_dEnergyMin = -1.0 * dYMax;
				m_dEnergyMax = 1.0 * dYMax;
				}
				
			//----- map Y values to [-60,60] range
			for( iState=0; iState<m_oModel.m_aStates.Count(); ++iState ) {
				pState0 = m_oModel.State(iState);
				pState0->m_dEnergyMin = 120.0f * ( pState0->m_dEnergyMin-m_dEnergyMin ) / (m_dEnergyMax-m_dEnergyMin) - 60.0f;
				pState0->m_dEnergyMax = 120.0f * ( pState0->m_dEnergyMax-m_dEnergyMin ) / (m_dEnergyMax-m_dEnergyMin) - 60.0f;
				}		
			for( iRate=0; iRate<m_oModel.m_aRates.Count(); ++iRate ) {
				pRate = m_oModel.Rate(iRate);
				pRate->m_dEnergyMin = 120.0f * ( pRate->m_dEnergyMin-m_dEnergyMin ) / (m_dEnergyMax-m_dEnergyMin) - 60.0f;
				pRate->m_dEnergyMax = 120.0f * ( pRate->m_dEnergyMax-m_dEnergyMin ) / (m_dEnergyMax-m_dEnergyMin) - 60.0f;
				}
			}

		//-- convert loaded qmf to energy landscape
		void BuildModel( ) { 
			
			//-- clear any previous contents
			m_aObj.DeleteAll();

			//-- add states 
			m_iStates=m_oModel.m_aStates.Count();
			for( int i=0; i<m_iStates; ++i ) { 
				CModelState * pState = (CModelState *)(m_oModel.m_aStates[i]);
				
				//-- Cube
				CGLBlock * pB;
				pB  = new CGLBlock();
				pB->SetBase( GLOB_BLOCK, GLfloat(pState->m_dXPos)-BLOCKSIZE/2.0f, 
							 GLfloat(pState->m_dEnergyMin)-BLOCKSIZE/2.0f, 
							 GLfloat(pState->m_dYPos)-BLOCKSIZE/2.0f, 
							 0.0, 0.0, 0.0,  BLOCKSIZE, BLOCKSIZE, BLOCKSIZE );
				COLORREF cr = g_aStateColors[min(pState->m_iClass,10)];
				COLORREF cr1 = RGB( GetRValue(cr)/2, GetGValue(cr)/2, GetBValue(cr)/2 );
				COLORREF cr2 = RGB( GetRValue(cr1)/2, GetGValue(cr1)/2, GetBValue(cr1)/2 );
				pB->SetColors( cr, cr, cr1, cr1, cr2, cr2 );
				
				AddObject( pB );

				char buffer[32];
				sprintf( buffer, "%d", i+1 );
				CGLText * pT;
				pT = new CGLText();
				pB->SetText( RGB(255,255,255), buffer );				
				}
			
			//-- add rates 
			CGLRibbon * pRibbon;
			CModelState * pState0, * pState1;
			CModelRate * pRate;
			m_iRates=m_oModel.m_aRates.Count();
			for( int i=0; i<m_iRates; ++i ) {
				pRate = (CModelRate *)(m_oModel.m_aRates[i]);
				pState0 = (CModelState *)(m_oModel.m_aStates[pRate->m_aiStates[0]]); 
				pState1 = (CModelState *)(m_oModel.m_aStates[pRate->m_aiStates[1]]); 
				pRibbon = new CGLRibbon();
				pRibbon->Set( 8, true,   
								GLfloat(pState0->m_dXPos), GLfloat(pState0->m_dEnergyMin), GLfloat(pState0->m_dYPos), 
								GLfloat(pRate->m_dEnergyMin), 
								GLfloat(pState1->m_dXPos), GLfloat(pState1->m_dEnergyMin), GLfloat(pState1->m_dYPos), 
								GLfloat(0.0),	
								RGB(0,0,255), RGB(255,0,0), RGB(0,0,255)
								);
				AddObject( pRibbon );
				}
			}			

		//-- convert loaded qmf to energy landscape
		void SetStimLevel( double dFrac ) { 
			//-- Set Y level for states 
			for( int i=0; i<m_iStates; ++i )
				((CGLBlock * ) m_aObj[i])->m_dY = GLfloat(((CModelState *)m_oModel.m_aStates[i])->Y(dFrac)) - BLOCKSIZE/2.0f;
			
			//-- Set Y level for rates 
			CModelRate * pRate;
			CModelState * pState0, * pState1;
			CGLRibbon * pRibbon;
			for( int i=0; i<m_iRates; ++i ) {
				pRate = (CModelRate *)(m_oModel.m_aRates[i]);
				pState0 = (CModelState *)(m_oModel.m_aStates[pRate->m_aiStates[0]]); 
				pState1 = (CModelState *)(m_oModel.m_aStates[pRate->m_aiStates[1]]); 
				pRibbon = (CGLRibbon *) m_aObj[i+m_iStates];
				
				pRibbon->m_dY = GLfloat(pState0->Y(dFrac));
				pRibbon->m_dYm = GLfloat(pRate->Y(dFrac))-pRibbon->m_dY;
				pRibbon->m_dY1 = GLfloat(pState1->Y(dFrac))-pRibbon->m_dY;
				}
			}			

		//----- Stimulus manipulation 
		int StimulusIncrement( int iStep=STIMULUS_INCREMENT ){ 
			m_iStimulus = min( 100, m_iStimulus+iStep );
			return m_iStimulus;
			}
		
		int StimulusDecrement( int iStep=STIMULUS_INCREMENT ){ 
			m_iStimulus = max( 0, m_iStimulus-iStep );
			return m_iStimulus;
			}

		int StimulusAdvance( bool bReset=false ) { 
			if( bReset ) {
				m_iStimulusDir=1;
				m_iStimulus=0;
				return 0;
				}
				
			m_iStimulus+=STIMULUS_INCREMENT*m_iStimulusDir;
			if( m_iStimulus >= 100 ) { 
				m_iStimulus=100;
				m_iStimulusDir=-1;
				}
			if( m_iStimulus <= 0 ) { 
				m_iStimulus=0;
				m_iStimulusDir=1;
				}
			Stimulus(-9999);
			return m_iStimulus;
			}
				
		int Stimulus( int iStimulus=-9999 ) { 
			if( iStimulus!=-9999 ) 
				m_iStimulus=max(0,min(100,iStimulus));
			return m_iStimulus;
			}
	};

//-----------------------------------------------------------------------------
class CVwEnergy : public CDialog { 
	public : 
	
		//-- scene global rotation 
		double m_dXRot;				
		double m_dYRot;
		double m_dZRot;
		
		//-- OpenGL scene info 
		HGLRC m_RC;						// OpenGL Rendering context
		CEnergyModel m_oEnergyModel;			// Scene container

		//-- Window handle 
		HWND m_hWnd;
		bool m_bSetup;

		//-- Controls for Top of window
		CDlgEdit * m_edtState0;
		CDlgEdit * m_edtScaling;
		CDlgEdit * m_edtMaxStim;
		CDlgCheck * m_chkGrid;
		CDlgCheck * m_chkStates;
		CDlgCheck * m_chkLegend;
		CDlgCheck * m_chkColor;
		//CDlgCheck * m_chkFine;
		CDlgCheck * m_chkYMax;
		CDlgEdit * m_edtYMax;

		CDlgControl * m_btnReset;
		CDlgCheck * m_chkStim;
		CDlgScroll * m_scrStimulus;
		CDlgCheck * m_chkRotate[3];

		CDlgControl * m_btnExport;
		CDlgControl * m_btnCopy;

		//-- openGL drawing area 
		int m_iWinHgt;
		RECT m_rctGL;
		int GLWidth() { 
			return m_rctGL.right-m_rctGL.left;
			}
		int GLHeight() { 
			return m_rctGL.bottom-m_rctGL.top;
			}

		//-- Set OpenGL pixel format ( for screen or printer )
		void GLSetupPixelFormat( HDC hdc, bool bBitmap ) {
			//-- setupPixelFormat
			PIXELFORMATDESCRIPTOR pfd;
			pfd.nSize = sizeof(pfd);						// size
			pfd.nVersion = 1;						        // version
			pfd.iPixelType = PFD_TYPE_RGBA;					// color type
			pfd.cColorBits = 24;							// prefered color depth
			pfd.cRedBits = 0; pfd.cRedShift = 0;			// color bits (ignored)
			pfd.cGreenBits = 0; pfd.cGreenShift = 0;
			pfd.cBlueBits = 0; pfd.cBlueShift = 0;
			pfd.cAlphaBits = 0; pfd.cAlphaShift = 0;		// no alpha buffer
			pfd.cAccumBits =  0;
			pfd.cAccumRedBits =  0;							// no accumulation buffer,
			pfd.cAccumGreenBits =  0;						// accum bits (ignored)
			pfd.cAccumBlueBits =  0;
			pfd.cAccumAlphaBits =  0;
			pfd.cDepthBits = 16;							// depth buffer
			pfd.cStencilBits = 0;							// no stencil buffer
			pfd.cAuxBuffers = 0;							// no auxiliary buffers
			pfd.iLayerType = PFD_MAIN_PLANE;				// main layer
			pfd.bReserved =  0;
			pfd.dwLayerMask =  0;
			pfd.dwVisibleMask =  0;
			pfd.dwDamageMask =  0;							// no layer, visible, damage masks */
			if( bBitmap ) { 
				pfd.dwFlags = PFD_SUPPORT_OPENGL | PFD_DRAW_TO_BITMAP;
				pfd.cRedBits = 8;
				pfd.cGreenBits = 8;
				pfd.cBlueBits = 8;
				}
			else 
				pfd.dwFlags = PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW | PFD_DOUBLEBUFFER;		
															// support double-buffering
						
			// get the best available match of pixel format for the device context  
			int iPixelFormat = ChoosePixelFormat(hdc, &pfd); 
			if( iPixelFormat==0 ) { 
				AddMessage( REPTYPE_ERROR, "OpenGL ChoosePixelFormat failed" );
				return;
				}
			// make that the pixel format of the device context 
			if( ! SetPixelFormat( hdc, iPixelFormat, &pfd ) ) {
				AddMessage( REPTYPE_ERROR, "OpenGL SetPixelFormat failed" );
				return;
				}
			}

		void GLInit( HDC hdc ) { 
			m_dYRot=m_dXRot=m_dZRot=0.0;
			m_RC = wglCreateContext( hdc );
			wglMakeCurrent( hdc, m_RC );
			}

		void GLInitDC( HDC hdc ) { 
			glMatrixMode( GL_PROJECTION );
			glLoadIdentity();
			//gluLookAt( 0.0, 0.0, -5.0, 0.0, 0.0, 5.0, 0.0, 1.0, 0.0 );
			glOrtho( -141.0, 141.0, -141.0, 141.0, -141.0, 141.0 );	
			//glFrustum( -141.0, 141.0, -141.0, 141.0, -141.0, 141.0 );	 

    		glMatrixMode( GL_MODELVIEW );
			glClearColor( 1.0, 1.0, 1.0, 1.0);		// Background Color
			glClearDepth(1.0);						// Depth Buffer Setup
			glEnable(GL_DEPTH_TEST);				// Enables Depth Testing
			glDepthFunc(GL_LESS);					// The Type Of Depth Test To Do
			}

		void GLDraw( HDC hdc, bool bBitmap ) { 
			m_oEnergyModel.GLFontSetup( hdc );
			//-- setup for drawing
			glEnable( GL_POINT_SMOOTH );
			glEnable( GL_LINE_SMOOTH );
			//glEnable( GL_BLEND );
			//glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
			//glPolygonMode( GL_FRONT, GL_FILL );
			
			glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
			glEnable(GL_NORMALIZE );
			glShadeModel( GL_SMOOTH );  //glShadeModel(GL_FLAT);
			glCullFace(GL_BACK);
			glEnable(GL_CULL_FACE);
			glLoadIdentity();

			glRotatef( GLfloat(m_dXRot), 1.0, 0.0, 0.0 );
			glRotatef( GLfloat(m_dYRot), 0.0, 1.0, 0.0 );
			glRotatef( GLfloat(m_dZRot), 0.0, 0.0, 1.0 );

			//----- The scene
			bool bFine = true;	//m_chkFine->GetCheck();
			if( bFine ) 
				m_oEnergyModel.SetFine(true);
			m_oEnergyModel.Render( );
			if( bFine ) 
				m_oEnergyModel.SetFine(false);

			if( m_chkGrid->GetCheck() ) { 
				double dY0 = -120.0f * m_oEnergyModel.m_dEnergyMin / (m_oEnergyModel.m_dEnergyMax-m_oEnergyModel.m_dEnergyMin) - 60.0f;

				//----- y=0 plane grid
				int i,j;
				glColor3f(0.75, 0.75, 0.75 );
				for( i=-100; i<=100; i+=20 ) {
					glBegin(GL_LINE_STRIP);
					for( j=-100; j<=100; j+=20 )
						glVertex3f( GLfloat(1.0f*i), GLfloat(dY0), GLfloat(1.0f*j) );
					glEnd();
					}
				for( j=-100; j<=100; j+=20 ) {
					glBegin(GL_LINE_STRIP);
					for( i=-100; i<=100; i+=20 )
						glVertex3f( GLfloat(1.0*i), GLfloat(dY0), GLfloat(1.0*j) );
					glEnd();
					}
				}
			glFlush();
			}
			
		//-----------------------------------------------------------------------
		double StimulusValue( ) { 
			return double( m_oEnergyModel.Stimulus()) /* * m_edtMaxStim->GetTextDouble() */ / 100.0f ;
			}

		//-----------------------------------------------------------------------
		void InitializeControls( ) {
			const int IC_HGT=20;
			const int IC_SEP=22;
			const int IC_CHWID=8;
			const int IC_TOP=10;

			//-- Text Labels
			int iLeft=10;
			CreateLabel( iLeft, IC_TOP+IC_SEP*0, 100, IC_HGT, IDC_STATIC, NULL, 0, SS_RIGHT )->SetText( "Base state:" );
			CreateLabel( iLeft, IC_TOP+IC_SEP*1, 100, IC_HGT, IDC_STATIC, NULL, 0, SS_RIGHT )->SetText( "Max Rate:" );
			CreateLabel( iLeft, IC_TOP+IC_SEP*2, 100, IC_HGT, IDC_STATIC, NULL, 0, SS_RIGHT )->SetText( "Max Stimulus:" );
			//(m_chkFine	= CreateCheckBox( iLeft, IC_TOP+IC_SEP*3, 100, IC_HGT, ID_CHKFINE, "Toggle fine grained drawing"))->SetText("Antialiased");
			(m_chkYMax	= CreateCheckBox( iLeft, IC_TOP+IC_SEP*3, 100, IC_HGT, ID_CHKYMAX, "Toggle Use of  Y axis limit"))->SetText("Set Y max =");
			m_chkYMax->SetCheck(true);
			
			iLeft+=110;
			m_edtState0	= CreateEdit( iLeft, IC_TOP+IC_SEP*0, 10+IC_CHWID*2 , IC_HGT, ID_EDTSTATE0 , "State with energy level = 0 (base)", ES_NUMBER );
			m_edtScaling= CreateEdit( iLeft, IC_TOP+IC_SEP*1, 10+IC_CHWID*12, IC_HGT, ID_EDTSCALING, "Set this value larger than the highest rate including stimulus", 0 );
			m_edtMaxStim= CreateEdit( iLeft, IC_TOP+IC_SEP*2, 10+IC_CHWID*8 , IC_HGT, ID_EDTMAXSTIM, "Maximum value of stimulus", 0 );
			m_edtYMax	= CreateEdit( iLeft, IC_TOP+IC_SEP*3, 10+IC_CHWID*8 , IC_HGT, ID_EDTYMAX , "Maximum Y value - scale limit" );

			m_edtState0->SetTextLimit( 2 );	
			m_edtScaling->SetTextLimit( 12 );
			m_edtMaxStim->SetTextLimit( 8 );
			m_edtYMax->SetTextLimit( 8 );

			#ifdef _DEBUG
			m_edtState0->SetText("2");	
			m_edtScaling->SetText("1e18");
			m_edtMaxStim->SetText("0.001");
			m_edtYMax->SetText("25");
			#else
			m_edtState0->SetText("1");	
			m_edtScaling->SetText("1000000");
			m_edtMaxStim->SetText("10.0");
			m_edtYMax->SetText("123456");
			#endif

			//-- option checkboxes
			iLeft+=110;
			(m_chkGrid	= CreateCheckBox( iLeft, IC_TOP+IC_SEP*0, 100, IC_HGT, ID_CHKGRID, "Toggle display of axis / grid")
							)->SetText("Grid");
			(m_chkStates= CreateCheckBox( iLeft, IC_TOP+IC_SEP*1, 100, IC_HGT, ID_CHKSTATES, "Toggle display of state markers")						
							)->SetText("States");
			(m_chkLegend= CreateCheckBox( iLeft, IC_TOP+IC_SEP*2, 100, IC_HGT, ID_CHKLEGEND, "Toggle display of energy scale bar")			
							)->SetText("Scale");
			(m_chkColor	= CreateCheckBox( iLeft, IC_TOP+IC_SEP*3, 100, IC_HGT, ID_CHKCOLOR, "Toggle display of color/monochrome")			
							)->SetText("Color");
			m_chkGrid->SetCheck(true);
			m_chkStates->SetCheck(true);
			m_chkLegend->SetCheck(true);
			m_chkColor->SetCheck(true);
			//m_chkFine->SetCheck(true);
									
			//-- Buttons
			iLeft+=110;
			(m_btnReset	= CreateButton( iLeft, IC_TOP+IC_SEP*0, 75, IC_HGT, ID_BTNRESET, NULL, "Reset view position")		
							)->SetText("Reset");
			(m_chkStim	= CreateCheckBox( iLeft, IC_TOP+IC_SEP*2, 75, IC_HGT, ID_CHKSTIMULUS, "Display movie mode of full stimulus range", NULL, BS_PUSHLIKE )		
							)->SetText("Stimulus");
			m_scrStimulus = CreateScrollBar( iLeft, IC_TOP+IC_SEP*3, 75, IC_HGT, ID_SCRSTIMULUS, "Stimulus level", SBS_HORZ );
			SetScrollRange( m_scrStimulus->GetHandle(), SB_CTL, 0, 100, TRUE );
							
			(m_chkRotate[0]	= CreateCheckBox( iLeft, IC_TOP+IC_SEP*1, 21, IC_HGT, ID_CHKROTATEX, "Display movie mode of rotation about X axis", NULL, BS_PUSHLIKE)
							)->SetText("X");
			iLeft+=27;							
			(m_chkRotate[1]	= CreateCheckBox( iLeft, IC_TOP+IC_SEP*1, 21, IC_HGT, ID_CHKROTATEY, "Display movie mode of rotation about Y axis", NULL, BS_PUSHLIKE)
							)->SetText("Y");
			iLeft+=27;							
			(m_chkRotate[2]	= CreateCheckBox( iLeft, IC_TOP+IC_SEP*1, 21, IC_HGT, ID_CHKROTATEZ, "Display movie mode of rotation about Z axis", NULL, BS_PUSHLIKE)
							)->SetText("Z");

			iLeft+=40;
			(m_btnExport= CreateButton( iLeft, IC_TOP+IC_SEP*0, 100, IC_HGT, ID_BTNEXPORT, NULL, "Export movie of full stimulus range")		
							)->SetText("Export Movie");
			(m_btnCopy	= CreateButton( iLeft, IC_TOP+IC_SEP*1, 100, IC_HGT, ID_BTNCOPYIMG, NULL, "Copy Image to clipboard")		
							)->SetText("Copy Image");
			
			}

		
		//-----------------------------------------------------------------------
		void CauseGLRedraw( ) { 			
			InvalidateRect( m_hWnd, &m_rctGL, false );
			}
			
		void CauseRedraw( ) { 			
			InvalidateRect( m_hWnd, NULL, false );
			}

		//-----------------------------------------------------------------------
		bool Setup( ) { 
			double dYMax = m_edtYMax->GetTextDouble();
			m_oEnergyModel.SetLevels( m_edtState0->GetTextInt()-1, 
								m_edtMaxStim->GetTextDouble(), 
								m_edtScaling->GetTextDouble(), 
								m_chkYMax->GetCheck() ? dYMax : -1.0 );
			m_oEnergyModel.BuildModel( );
			CauseGLRedraw();
			m_bSetup=true;
			
			return true;
			}
			
		//-----------------------------------------------------------------------
		void FileOpen( const char * pchFile ) { 
			if( m_oEnergyModel.m_oModel.OpenQmf( pchFile ) ) 
				Setup();
			}
			
		//-----------------------------------------------------------------------
		void ModelOpen( CModel * pModel ) { 
			m_oEnergyModel.m_oModel.Copy( *pModel );

			// initial setup to get max energy level default value
			m_oEnergyModel.SetLevels( m_edtState0->GetTextInt()-1, 
								m_edtMaxStim->GetTextDouble(), 
								m_edtScaling->GetTextDouble(), 
								-1.0 );
			m_edtYMax->SetTextDouble( m_oEnergyModel.m_dEnergyMax );
			Setup();
			}
			
		//-----------------------------------------------------------------------
		void Draw( HDC hdc ) {
			//-- Open GL drawing 
			wglMakeCurrent( hdc, m_RC);
			glViewport( m_rctGL.left, m_iWinHgt-m_rctGL.bottom, GLWidth(), GLHeight() );			

			//-- Draw the GL window 
			GLDraw( hdc, false );
			SwapBuffers( hdc );

			//-- GDI drawing
			HPEN penRed = CreatePen( 0, 0, RGB(255,0,0) );
			HPEN penGray = CreatePen( 0, 0, RGB(64,64,64) );
			HPEN penBlack = CreatePen( 0, 0, RGB(0,0,0) );
			HPEN penBlue = CreatePen( 0, 0, RGB(0,0,255) );
			HPEN penWas = (HPEN) SelectObject( hdc, penBlue );
			
			HBRUSH brWhite = (HBRUSH) GetStockObject(WHITE_BRUSH);
			HBRUSH brGray = (HBRUSH) GetStockObject(LTGRAY_BRUSH);
			HBRUSH brWas = (HBRUSH) SelectObject( hdc, brGray );
			
			//  Frame around the GL window 
			MoveToEx( hdc, m_rctGL.left-1, m_rctGL.top-1, NULL );
			LineTo( hdc, m_rctGL.right+1, m_rctGL.top-1 );
			LineTo( hdc, m_rctGL.right+1, m_rctGL.bottom+1 );
			LineTo( hdc, m_rctGL.left-1, m_rctGL.bottom+1 );
			LineTo( hdc, m_rctGL.left-1, m_rctGL.top-1 );
			
			//-- Cleanup 
			SelectObject( hdc, brWas );
			//DeleteObject( brWhite );		// unneccesary but not harmful
			//DeleteObject( brGray );		// unneccesary but not harmful
			SelectObject( hdc, penWas );
			DeleteObject( penRed );
			DeleteObject( penGray );
			DeleteObject( penBlack );
			DeleteObject( penBlue );
			}		
			
			
		//-----------------------------------------------------------------------
		BOOL OnKey( DWORD dwKey ) { 
			switch( dwKey ) { 
				case 37 :				// left
					m_dYRot-=ROTATEANGLE;
					break;
				case 39 :				// right
					m_dYRot+=ROTATEANGLE;
					break;
				case 38 :				// Up
					m_dXRot-=ROTATEANGLE;
					break;
				case 40 :				// down
					m_dXRot+=ROTATEANGLE;
					break;
				case 36 :				// Home (CW)
					m_dZRot-=ROTATEANGLE;
					break;
				case 35 :				// End (CCW)
					m_dZRot+=ROTATEANGLE;
					break;
				case 45 :				// - : decrease stimulus
				case 189 :
					m_oEnergyModel.StimulusDecrement( );
					m_oEnergyModel.SetStimLevel( StimulusValue() );
					break;
				case 43 :				// + : increase stimulus
				case 187 :
					m_oEnergyModel.StimulusIncrement( );
					m_oEnergyModel.SetStimLevel( StimulusValue() );
					break;
					
				default : 
					/*
					#ifdef _DEBUG
						char buffer[100];
						sprintf( buffer, "ED Key : %d\n", int(dwKey) );
						OutputDebugString(  buffer );
					#endif
					*/
					return FALSE;
				}
						
			CauseGLRedraw();	
			return TRUE;
			} 		
			
		void OnCheckStates() { 
			bool bCk = m_chkStates->GetCheck();
			for( int i=0; i<m_oEnergyModel.m_iStates; ++i ) 
				((CGLObject *)(m_oEnergyModel.m_aObj[i]))->m_bVisible=bCk;
			CauseGLRedraw();	
			}
			
		void OnCheck() { 
			CauseGLRedraw();	
			}
			
		//-----------------------------------------------------------------------
		void OnSetSize( int iWid, int iHgt ) { 
			m_iWinHgt=iHgt;
			SetRect( &m_rctGL, GLWINDOWMARGIN+1, TOPOFGLWINDOW, 
						max(GLWINDOWMARGIN+30, iWid-GLWINDOWMARGIN-1 ), 
						max(TOPOFGLWINDOW+30, iHgt-GLWINDOWMARGIN-1) );
			glViewport( m_rctGL.left, iHgt-m_rctGL.bottom, GLWidth(), GLHeight() );			
			CauseRedraw();
			}
			
		//-----------------------------------------------------------------------
		void OnBtnReset() {
			m_dXRot=0.0;
			m_dYRot=0.0;
			m_dZRot=0.0;
			CauseRedraw();
			SetFocus( m_hWnd );
			}
			
		void OnBtnStimulus(){
			bool bSet = m_chkStim->GetCheck();
			if( bSet ) {
				SetTimer( m_hWnd, TIMER_STIMULUS, 50, NULL );
				m_oEnergyModel.StimulusAdvance( TRUE );
				}
			else 
				KillTimer( m_hWnd, TIMER_STIMULUS );
			SetFocus( m_hWnd );
			}	

		void GLDrawBitmap( BYTE * pMem, int iWid, int iHgt ) { 
			//-- create bitmap DC
			HDC hdcWin = GetDC( m_hWnd );
			HDC hdcBmp = CreateCompatibleDC( hdcWin );
			
			//----- Prepare bitmap for Open GL rendering
			BITMAPINFO info;
			info.bmiHeader.biSize = sizeof(info.bmiHeader);
			info.bmiHeader.biWidth = iWid;
			info.bmiHeader.biHeight = iHgt;
			info.bmiHeader.biPlanes = 1;
			info.bmiHeader.biBitCount = 24;
			info.bmiHeader.biCompression = BI_RGB;
			info.bmiHeader.biClrImportant = 0;
			info.bmiHeader.biClrUsed = 0;
			info.bmiHeader.biSizeImage = 0;
			info.bmiHeader.biXPelsPerMeter = 0;
			info.bmiHeader.biYPelsPerMeter = 0;
			info.bmiColors->rgbBlue=192;
			info.bmiColors->rgbGreen=192;
			info.bmiColors->rgbRed=192;

			BYTE * pBits;
			HBITMAP hBmp = CreateDIBSection( hdcBmp, &info, DIB_RGB_COLORS, (void **)(&pBits), NULL, 0);
			SelectObject( hdcBmp, hBmp );

			GLSetupPixelFormat( hdcBmp, true );
			HGLRC rc = wglCreateContext( hdcBmp );
			wglMakeCurrent( hdcBmp, rc );
			GLInitDC( hdcBmp );

			//----- Render to bitmap
			SelectObject(  hdcBmp, GetStockObject(DEFAULT_GUI_FONT) );
			GLYPHMETRICSFLOAT agmf['z'-'0'+1];	
			wglUseFontOutlines( hdcBmp, '0', 'z'-'0', 1000, 0.1f, 0.1f, WGL_FONT_POLYGONS, agmf);
			GLDraw( hdcBmp, true );

			memcpy( pMem, &info.bmiHeader, sizeof( BITMAPINFOHEADER ) );
			memcpy( pMem+sizeof( BITMAPINFOHEADER ), pBits, iWid*iHgt*3 );
			
			//---- cleanup
			wglDeleteContext( rc );
			DeleteObject( hBmp );
			DeleteDC( hdcBmp );
			ReleaseDC( m_hWnd, hdcWin );
			SetFocus( m_hWnd );
			}
			
		void OnBtnCopyImage(){
			int iWid=m_rctGL.right-m_rctGL.left;
			int iHgt=m_rctGL.bottom-m_rctGL.top;
			//----- Create memory object 
			HGLOBAL hBits = GlobalAlloc( GHND, sizeof(BITMAPINFOHEADER)+iWid*iHgt*3+1 );
			
			//----- Fill it with bits
			BYTE * pMem = (BYTE *)GlobalLock( hBits );
			GLDrawBitmap( pMem, iWid, iHgt );
			GlobalUnlock( pMem );

			//----- Copy to clipboard
			if( OpenClipboard( m_hWnd ) ) { 
				EmptyClipboard();
				if( NULL==SetClipboardData( CF_DIB, hBits ) ) 
					GlobalFree( hBits );
				CloseClipboard( );
				}
			else 
				GlobalFree( hBits );
			
			}	
			
		void OnBtnRotate(int iAxis){
			bool bSet = m_chkRotate[0]->GetCheck() || m_chkRotate[1]->GetCheck() || m_chkRotate[2]->GetCheck() ;
			if( bSet )
				SetTimer( m_hWnd, TIMER_ROTATE, 50, NULL );
			else 
				KillTimer( m_hWnd, TIMER_ROTATE );
			SetFocus( m_hWnd );
			}	
			
		void OnBtnExport(){
			const int FRAMECOUNT = 64;
			const int BASECHARID = 1;
			const int BASEFRAMEID = BASECHARID + FRAMECOUNT + 1;
			double dSave = m_dYRot;
			
			//----- Bitmap storage 
			int iWid=m_rctGL.right-m_rctGL.left;
			iWid=(iWid>>2)<<2;
			int iHgt=m_rctGL.bottom-m_rctGL.top;
			BYTE * p = new BYTE[ sizeof(BITMAPINFOHEADER)+iWid*iHgt*3+1 ];

			//-- select the file name to save to 
			char achLongName[_MAX_PATH]={0};
			char achFileName[_MAX_PATH]={0};

			if( ! GetSaveFileName( achLongName, 
							"Shockwave file(*.swf)\0*.swf\0\0", 
							"swf", GetParent(m_hDlg), achFileName ) ) 
				return;

			//----- .SWF create
			int iCharID=0;
			int iPlaceID=FRAMECOUNT+1;	
			int iFrames=0;
			CSWF oSWF;
			oSWF.Create( achLongName, iWid, iHgt, 10, FRAMECOUNT );

			//----- Step through frames - add bitmaps to file 
			// TODO - detemine which are changing and alter them 
			double dPrevStim=StimulusValue();
			
			for( int iFrame=0; iFrame<FRAMECOUNT; ++iFrame ) { 
				if( m_chkStim->GetCheck() ) 
					m_oEnergyModel.SetStimLevel( double(iFrame)/FRAMECOUNT );
				if( m_chkRotate[0]->GetCheck() ) 
					m_dXRot = 360.0 * double(iFrame) / FRAMECOUNT;
				if( m_chkRotate[1]->GetCheck() ) 
					m_dYRot = 360.0 * double(iFrame) / FRAMECOUNT;
				if( m_chkRotate[2]->GetCheck() ) 
					m_dZRot = 360.0 * double(iFrame) / FRAMECOUNT;
				GLDrawBitmap( p, iWid, iHgt );	// Fill in *p
				oSWF.BMPCopy( BASECHARID+iFrames, BASEFRAMEID+iFrames, iHgt, iWid, p+sizeof(BITMAPINFOHEADER) );
				++iFrames;
				oSWF.Flush();
				}
				
			m_oEnergyModel.SetStimLevel( dPrevStim );
			
			//---- Add timeline 
			for( int iFrame=0; iFrame<iFrames; ++iFrame ) 
				oSWF.AddFrame( BASEFRAMEID+iFrame );
			
			//---- Cleanup 
			oSWF.Close();
			m_dYRot = dSave;
			delete [] p;
			SetFocus( m_hWnd );
			}	
			
		//-----------------------------------------------------------------------
		void OnRotateIncrement() { 
			if( m_chkRotate[0]->GetCheck() )
				if( (m_dXRot += ROTATEANGLE) > 360.0 )
					m_dXRot -= 360.0;
				
			if( m_chkRotate[1]->GetCheck() )
				if( (m_dYRot += ROTATEANGLE) > 360.0 ) 
					m_dYRot -= 360.0;
			if( m_chkRotate[2]->GetCheck() )
				if( (m_dZRot += ROTATEANGLE) > 360.0 ) 
					m_dZRot -= 360.0;
					
			CauseGLRedraw();
			}
			
		//-----------------------------------------------------------------------
		void OnStimulusIncrement() { 
			m_scrStimulus->SetPos( m_oEnergyModel.StimulusAdvance() );
			m_oEnergyModel.SetStimLevel( StimulusValue() );
			CauseGLRedraw();
			}
			
		void OnScroll( DWORD iOp, DWORD dwPos ) { 
			int iPos=m_scrStimulus->GetPos();
			int iPosSave=iPos;
			
			switch( iOp ) {
				case SB_LEFT :
					iPos=max(0,m_scrStimulus->GetPos()-5);
					break;
				case SB_RIGHT :
					iPos=min(100,m_scrStimulus->GetPos()+5);
					break;
				case SB_LINELEFT :
					iPos=max(0,m_scrStimulus->GetPos()-1);
					break;
				case SB_LINERIGHT :
					iPos=min(100,m_scrStimulus->GetPos()+1);
					break;
				case SB_PAGELEFT :
					iPos=max(0,iPos-20);
					break;
				case SB_PAGERIGHT :
					iPos=min(100,iPos+20);
					break;
				case SB_THUMBPOSITION :
					iPos=dwPos;
					break;
				case SB_THUMBTRACK :
					break;
				default : 
					return;
				}

			m_scrStimulus->SetPos( iPos );
			m_oEnergyModel.Stimulus( iPos );
			m_oEnergyModel.SetStimLevel( StimulusValue() );
			CauseGLRedraw();
			}

	};

//-----	Child Window - base window in MDI which has a dialog superimposed on it
LRESULT CALLBACK EnergyWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	PAINTSTRUCT ps;
	CVwEnergy * pVw;
	HDC hdc;
	switch( message ) {
		case WM_CREATE : 
			pVw = new CVwEnergy;
			pVw->m_bSetup=false;
			pVw->m_hDlg=hWnd; 
			pVw->m_hWnd=hWnd;
			SetWindowLong( hWnd, 0,  (long) pVw ) ;
			pVw->m_iWinHgt=100;
			pVw->InitializeControls();
			hdc=GetDC( hWnd );
			pVw->GLSetupPixelFormat( hdc, false );
			pVw->GLInit( hdc );
			pVw->GLInitDC( hdc );
			glViewport( GLWINDOWMARGIN+1, GLWINDOWMARGIN+1, 10, 10);
			ReleaseDC( hWnd, hdc );
			return 0;		

		case WM_MODEL_OPEN : 
			if( wParam==0 ) {
				pVw = (CVwEnergy *)GetWindowLong( hWnd, 0 );
				pVw->FileOpen( (char *) lParam );
				}
				
			if( wParam==1 ) { 
				pVw = (CVwEnergy *)GetWindowLong( hWnd, 0 );
				pVw->ModelOpen( (CModel *) lParam );
				}
				
			return 0;
						
		case WM_COMMAND : 
			pVw = (CVwEnergy *)GetWindowLong( hWnd, 0 );
			switch( wParam ) { 
				case ID_CHKYMAX : 
				case (EN_CHANGE<<16)|ID_EDTSTATE0 : 
				case (EN_CHANGE<<16)|ID_EDTSCALING : 
				case (EN_CHANGE<<16)|ID_EDTMAXSTIM : 
				case (EN_CHANGE<<16)|ID_EDTYMAX : 
					if( pVw->m_bSetup )	// if the edits change and the model has been setup, redo setup.
						pVw->Setup();
					break;
					
				case ID_CHKCOLOR :
					pVw->m_oEnergyModel.SetColor( pVw->m_chkColor->GetCheck() );
				case ID_CHKGRID :
				case ID_CHKFINE : 
				case ID_CHKLEGEND :
					pVw->OnCheck();
					break;
					
				case ID_CHKSTATES :
					pVw->OnCheckStates();
					break;

				case ID_CHKSTIMULUS	:
					pVw->OnBtnStimulus();
					break;

				case ID_BTNRESET	:	
					pVw->OnBtnReset();
					break;

				case ID_BTNCOPYIMG	:
					pVw->OnBtnCopyImage();
					break;

				case ID_BTNEXPORT	:
					pVw->OnBtnExport();
					break;

				case ID_CHKROTATEX	:	
					pVw->OnBtnRotate(1);
					break;

				case ID_CHKROTATEY	:	
					pVw->OnBtnRotate(2);
					break;

				case ID_CHKROTATEZ	:	
					pVw->OnBtnRotate(3);
					break;
				}	
			return 0;

		case WM_MDIACTIVATE :		// Change the menu when this window is activated.
			return 0;		
			
		case WM_KEYDOWN : 
			pVw = (CVwEnergy *)GetWindowLong( hWnd, 0 );
			pVw->OnKey( wParam );
			return 0;

		case WM_HSCROLL : 
			pVw = (CVwEnergy *)GetWindowLong( hWnd, 0 );
			pVw->OnScroll( LOWORD(wParam), HIWORD(wParam) ); 
			//pVw->Setup();
			return 0;
			
		case WM_PAINT : {
			HDC hdc = BeginPaint(hWnd, &ps);

			//----- Manual Drawing 
			pVw = (CVwEnergy *) GetWindowLong(  hWnd, 0 );
			pVw->Draw( hdc );
			
			//----- Done
			EndPaint(hWnd, &ps);
			}
			return 0;

		case WM_SIZE : 
			pVw = (CVwEnergy *)GetWindowLong( hWnd, 0 );
			pVw->OnSetSize( LOWORD(lParam), HIWORD(lParam) );
			break;

		case WM_DESTROY : 
			pVw = (CVwEnergy *)GetWindowLong( hWnd, 0 );
			delete pVw;
			SetWindowLong( hWnd, 0,  0 ) ;
			return 0;	

		case WM_TIMER : 
			pVw = (CVwEnergy *)GetWindowLong( hWnd, 0 );
			if( wParam == TIMER_ROTATE ) 
				pVw->OnRotateIncrement();
			else if( wParam == TIMER_STIMULUS ) 
				pVw->OnStimulusIncrement();
			return 0;
				
		case WM_LBUTTONDOWN : 
			pVw = (CVwEnergy *)GetWindowLong( hWnd, 0 );
			if( HIWORD(lParam) >= pVw->m_rctGL.top && HIWORD(lParam) <= pVw->m_rctGL.bottom 
				&& LOWORD(lParam) >= pVw->m_rctGL.left && LOWORD(lParam) <= pVw->m_rctGL.right ) { 
				SetFocus( pVw->m_hWnd );
				SetCapture( pVw->m_hWnd );
				g_bDrag=true;
				g_iXDrag=LOWORD(lParam);
				g_iYDrag=HIWORD(lParam);
				}
			return 0;
			
		case WM_MOUSEMOVE :			//-- Mouse move during drag operation
			if( g_bDrag ) { 
				pVw = (CVwEnergy *)GetWindowLong( hWnd, 0 );
				int iX=LOWORD(lParam);
				int iY=HIWORD(lParam);
				pVw->m_dYRot += 0.2*double(iX-g_iXDrag);
				pVw->m_dXRot += 0.33*double(iY-g_iYDrag);
				g_iXDrag=iX;
				g_iYDrag=iY;
				pVw->CauseGLRedraw();
				}
			return 0;
			
		case WM_LBUTTONUP :			//-- end of drag operation
			if( g_bDrag ) { 
				SetCapture( NULL );
				g_bDrag=false;
				}
			return 0;
			
		case WM_GETMINMAXINFO : 
			((MINMAXINFO*)lParam)->ptMinTrackSize.x=690;
			((MINMAXINFO*)lParam)->ptMinTrackSize.y=275;
			return 0;

		}
		
	return DefMDIChildProc( hWnd, message, wParam, lParam);
	}

